/* Copyright 2016 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

#ifndef TENSORFLOW_SERVING_CORE_SERVABLE_HANDLE_H_
#define TENSORFLOW_SERVING_CORE_SERVABLE_HANDLE_H_

#include <algorithm>
#include <cstddef>
#include <memory>
#include <type_traits>

#include "tensorflow_serving/core/loader.h"
#include "tensorflow_serving/util/any_ptr.h"

namespace tensorflow {
namespace serving {

/// A non-templatized handle to a servable, used internally in the
/// Manager to retrieve a type-erased servable object from the Loader.
/// The handle keeps the underlying object alive as long as the handle is alive.
/// The frontend should not hold onto it for a long time, because holding it can
/// delay servable reloading.
class UntypedServableHandle {
 public:
  virtual ~UntypedServableHandle() = default;

  virtual const ServableId& id() const = 0;

  virtual AnyPtr servable() = 0;
};

/// A smart pointer to the underlying servable object T retrieved from the
/// Loader. Frontend code gets these handles from the ServableManager. The
/// handle keeps the underlying object alive as long as the handle is alive. The
/// frontend should not hold onto it for a long time, because holding it can
/// delay servable reloading.
///
/// The T returned from the handle is generally shared among multiple requests,
/// which means any mutating changes made to T must preserve correctness
/// vis-a-vis the application logic. Moreover, in the presence of multiple
/// request threads, thread-safe usage of T must be ensured.
///
/// T is expected to be a value type, and is internally stored as a pointer.
/// Using a pointer type for T will fail to compile, since it would be a mistake
/// to do so in most situations.
///
/// Example use:
///
///     // Define or use an existing servable:
///     class MyServable {
///     public:
///       void MyMethod();
///     };
///
///     // Get your handle from a manager.
///     ServableHandle&lt;MyServable> handle;
///     TF_RETURN_IF_ERROR(manager->GetServableHandle(id, &handle));
///
///     // Use your handle as a smart-pointer:
///     handle->MyMethod();
///
template <typename T>
class ServableHandle {
 public:
  static_assert(!std::is_pointer<T>::value,
                "Servables are implicitly passed as pointers, please use T "
                "instead of T*.");

  /// ServableHandle is null by default.
  ServableHandle() = default;

  const ServableId& id() const { return untyped_handle_->id(); }

  // Smart pointer operations.

  T& operator*() const { return *get(); }

  T* operator->() const { return get(); }

  T* get() const { return servable_; }

  operator bool() const { return get() != nullptr; }

  // See the end of this file for comparison operators, which must be declared
  // at namespace scope to support left-hand-side arguments of different types.

 private:
  friend class Manager;

  explicit ServableHandle(std::unique_ptr<UntypedServableHandle> untyped_handle)
      : untyped_handle_(std::move(untyped_handle)),
        servable_(untyped_handle_ == nullptr
                      ? nullptr
                      : untyped_handle_->servable().get<T>()) {}

  std::unique_ptr<UntypedServableHandle> untyped_handle_;
  T* servable_ = nullptr;
};

/// An implementation of UntypedServableHandle using shared_ptr to do
/// ref-counting on the Loader that owns the Servable.
class SharedPtrHandle final : public UntypedServableHandle {
 public:
  ~SharedPtrHandle() override = default;

  explicit SharedPtrHandle(const ServableId& id, std::shared_ptr<Loader> loader)
      : id_(id), loader_(std::move(loader)) {}

  AnyPtr servable() override { return loader_->servable(); }

  const ServableId& id() const override { return id_; }

 private:
  const ServableId id_;
  std::shared_ptr<Loader> loader_;
};

// We compare handles using both the servable pointer and the id. So if you
// share the same pointer amongst different versions, the handles won't be
// equal.
template <typename T, typename U>
constexpr bool operator==(const ServableHandle<T>& l,
                          const ServableHandle<U>& r) {
  return l.get() == r.get() && l.id() == r.id();
}

template <typename T, typename U>
constexpr bool operator!=(const ServableHandle<T>& l,
                          const ServableHandle<U>& r) {
  return !(l == r);
}

}  // namespace serving
}  // namespace tensorflow

#endif  // TENSORFLOW_SERVING_CORE_SERVABLE_HANDLE_H_
